/*==============================================================================
    FILE:           user_audio_w_cdc.c

    DESCRIPTION:    Implements a linked list used to manage audio buffers passed between USB
                    and the CODECs.

    Copyright (c) 2018 Closed Loop Design, LLC

    This software is supplied "AS IS" without any warranties, express, implied
    or statutory, including but not limited to the implied warranties of fitness
    for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
    extends you a royalty-free right to use, reproduce and distribute this source
    file as well as executable files created using this source file for use with 
    Analog Devices SC5xx family processors only. Nothing else gives you 
    the right to use this source file.


==============================================================================*/
/*!
 * @file      ll_mgr.c
 * @brief     Audio buffer linked list manager.
 *
 * @details
 *            Implements a linked list used to manage audio buffers passed between USB
 *            and the CODECs.
 *
 *            Copyright (c) 2018 Closed Loop Design, LLC
 *
 *            This software is supplied "AS IS" without any warranties, express, implied
 *            or statutory, including but not limited to the implied warranties of fitness
 *            for purpose, satisfactory quality and non-infringement. Closed Loop Design LLC
 *            extends you a royalty-free right to use, reproduce and distribute this source
 *            file as well as executable files created using this source file for use with 
 *            Analog Devices SC5xx family processors only. Nothing else gives you 
 *            the right to use this source file.
 *
 */

#include "ll_mgr.h"
#include "cld_sc58x_audio_2_0_w_cdc_lib.h"
#include <runtime/int/interrupt.h>
#include <stdio.h>

#define LL_MGR_TOTAL_NUM_NODES      16

static LL_Mgr_Data ll_mgr_free_nodes;
static LL_Mgr_Node ll_mgr_nodes[LL_MGR_TOTAL_NUM_NODES] __attribute__ ((section(".l3_uncached_data")));

/**
 * Initializes the audio buffer linked list.
 *
 * @param p_ll - pointer to the linked list being initialized.
 */
void ll_mgr_init (LL_Mgr_Data * p_ll)
{
    static unsigned char first_call = 0;
    unsigned int i;

    /* If it is the first time this function has been called initialize the free nodes linked list. */
    if (first_call == 0)
    {
        first_call = 1;

        ll_mgr_free_nodes.num_elements = 0;
        for (i = 0; i < LL_MGR_TOTAL_NUM_NODES; i++)
        {
            ll_mgr_insert (&ll_mgr_free_nodes, &ll_mgr_nodes[i]);
        }
    }
    p_ll->num_elements = 0;
    p_ll->p_head = 0;
    p_ll->p_tail = 0;
}

/**
 * Inserts the specified data into the audio buffer linked list addressed by p_ll.
 *
 * @param p_ll - pointer to the linked list being modified.
 * @param p_node - pointer to data being inserted.
 */
void ll_mgr_insert (LL_Mgr_Data * p_ll, LL_Mgr_Node * p_node)
{
    adi_rtl_disable_interrupts();
    if (p_ll->p_head == 0)
    {
        p_ll->p_head = p_node;
        p_node->p_next = 0;
        p_node->p_prev = 0;
        p_ll->p_tail = p_node;
    }
    else
    {
        p_node->p_next = p_ll->p_head;
        p_ll->p_head->p_prev = p_node;
        p_node->p_prev = 0;
        p_ll->p_head = p_node;
    }
    p_ll->num_elements++;
    adi_rtl_reenable_interrupts();
}

/**
 * Removes data from the tail of the audio buffer linked list addressed by p_ll.
 *
 * @param p_ll - pointer to the linked list being modified.
 * @retval the linked list node removed from the tail of the list.
 */
LL_Mgr_Node * ll_mgr_remove (LL_Mgr_Data * p_ll)
{
    LL_Mgr_Node * rv;

    adi_rtl_disable_interrupts();
    rv = p_ll->p_tail;

    if (rv)
    {
        if (rv == p_ll->p_head)
        {
            p_ll->p_head = p_ll->p_head->p_next;
            if (p_ll->p_head != 0)
            {
                p_ll->p_head->p_prev = 0;
            }
            else
            {
                p_ll->p_tail = 0;
            }
            rv->p_next = 0;
        }
        else if (rv == p_ll->p_tail)
        {
            p_ll->p_tail = p_ll->p_tail->p_prev;
            p_ll->p_tail->p_next = 0;
            rv->p_prev = 0;
        }
        else
        {
            rv->p_prev->p_next = rv->p_next;
            rv->p_next->p_prev = rv->p_prev;
            rv->p_prev = 0;
            rv->p_next = 0;
        }
        p_ll->num_elements--;
    }
    else
    {
    }
    adi_rtl_reenable_interrupts();
    return rv;
}

/**
 * Adds a node to the free node linked list
 *
* @param p_node - pointer to node being freed.
 */
void ll_mgr_free_node (LL_Mgr_Node * p_node)
{
    ll_mgr_insert (&ll_mgr_free_nodes, p_node);
}

/**
 * Clears the nodes from the specified linked list.
 *
 * @param p_ll - pointer to the linked list being modified.
 */
void ll_mgr_reset (LL_Mgr_Data * p_ll)
{
    LL_Mgr_Node * tmp;
    unsigned int i;
    unsigned long num_elements = p_ll->num_elements;

    for (i = 0; i < num_elements; i++)
    {
        tmp = ll_mgr_remove(p_ll);
        ll_mgr_free_node(tmp);
    }
}

/**
 * Returns a node from the free node list.
 *
 * @param clear_data - When not 0 the node's data buffer is initialized with 0.
 * @retval the linked list node removed from the free node list.
 */
LL_Mgr_Node * ll_mgr_get_node (unsigned char clear_data)
{
    unsigned int i;

    LL_Mgr_Node * rv = ll_mgr_remove(&ll_mgr_free_nodes);
    if (rv)
    {
        if (clear_data)
        {
            for (i = 0; i < LL_MGR_AUDIO_DATA_SIZE; i++)
            {
                rv->data[i] = 0;
            }
        }
    }
    else
    {
        printf("No free ll nodes\n\r");
    }
    return rv;
}
